import json


INPUT_MESSAGE = "INPUT_MESSAGE=>"
INTERRUPT_MESSAGE = "INTERRUPT_MESSAGE=>"
_OUTPUT_MESSAGE = "OUTPUT_MESSAGE=>"
_INFOMATION_MESSAGE = "INFOMATION_MESSAGE=>"
_LOG = "LOG=>"


CHAT_DATA = {"output": "", "info": ""}


def log(content, *args, end="\n", **kwargs):
    print(_LOG + content, *args, end=end, **kwargs)


def clear_chat_data():
    CHAT_DATA["output"] = ""
    CHAT_DATA["info"] = ""


def output(role, content):
    assert role == "assistant"
    CHAT_DATA["output"] = content
    print(
        _OUTPUT_MESSAGE
        + json.dumps({"role": role, "content": content}, ensure_ascii=False)
    )


def info(role, content):
    CHAT_DATA["info"] += content
    print(
        _INFOMATION_MESSAGE
        + json.dumps({"role": role, "content": content}, ensure_ascii=False)
    )



import os
import re
import subprocess
import fnmatch
from typing import Dict, Tuple
from .message import info
from .lib import get_formatted_filelist_str, read_local_file, resolve_relative_path
from pycoze import ai
from pycoze.api import window, web
import traceback
import json


class InvalidToolError(Exception):
    """无效工具错误"""

    pass


class Tool:
    """工具基类"""

    def __init__(self, params: Dict[str, str], cwd: str):
        self.params = params
        self.cwd = cwd

    def validate(self) -> bool:
        """验证参数是否有效"""
        raise NotImplementedError

    def execute(self) -> str:
        """执行工具并返回结果"""
        raise NotImplementedError


class ExecuteCommandTool(Tool):
    """执行命令工具"""

    def validate(self) -> bool:
        return "command" in self.params and "requires_approval" in self.params

    def execute(self) -> str:
        try:
            if self.params["requires_approval"]:
                approve = window.confirm(
                    "confirm",
                    f"Do you want to execute the command{self.params['command']}? ",
                )
                if not approve:
                    return "User does not wish to run the command."
            result = subprocess.run(
                self.params["command"],
                shell=True,
                capture_output=True,
                text=True,
                cwd=self.cwd,
            )
            if result.returncode != 0:
                raise subprocess.CalledProcessError(
                    result.returncode,
                    self.params["command"],
                    result.stdout,
                    result.stderr,
                )
            return result.stdout + result.stderr
        except subprocess.CalledProcessError as e:
            return f"命令执行失败: {e.stderr}"
        except Exception as e:
            return f"执行命令时发生错误: {str(e)}"


class ReadFileTool(Tool):
    """读取文件工具"""

    def validate(self) -> bool:
        return "path" in self.params and os.path.exists(
            resolve_relative_path(self.cwd, self.params["path"])
        )

    def execute(self) -> str:
        path = resolve_relative_path(self.cwd, self.params["path"])
        content = read_local_file(path)
        return f"Successfully read the file, the content of the file [[{path}]] is:\n{content}\n\n\n"


class ReadMultipleFilesTool(Tool):
    """读取多个文件工具"""

    def validate(self) -> bool:
        return "paths_list" in self.params

    def execute(self) -> str:
        return_str = ""
        for path in self.params["paths_list"]:
            if os.path.exists(path):
                path = resolve_relative_path(self.cwd, path)
                content = read_local_file(path)
                return_str += f"The content of the file [[{path}]] is:\n{content}\n\n\n"
            else:
                return_str += f"File {path}  does not exist.\n\n\n"
        if return_str == "":
            return_str = "No files were read"
        return "Result of reading multiple files:" + return_str


class WriteFileTool(Tool):
    """写入或覆盖文件工具"""

    def validate(self) -> bool:
        return "path" in self.params and "content" in self.params

    def execute(self) -> str:
        path = resolve_relative_path(self.cwd, self.params["path"])
        # 确保路径是绝对路径
        if not os.path.isabs(path):
            path = os.path.join(os.getcwd(), path)
        # 创建目录并写入或覆盖文件
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w", encoding="utf-8") as f:
            f.write(self.params["content"])
        return f"File [[{path}]] written successfully."


class ReplaceInFileTool(Tool):
    """替换文件部分内容工具"""

    def validate(self) -> bool:
        return "path" in self.params and "diff" in self.params

    def execute(self) -> str:
        path = resolve_relative_path(self.cwd, self.params["path"])
        with open(path, "r", encoding="utf-8") as f:
            content = f.read()

        # 处理差异内容
        diff_content = self.params["diff"]
        for diff_block in diff_content.split("<<<<<<< SEARCH")[1:]:
            search, replace = diff_block.split("=======")
            search = search.strip()
            replace = replace.split(">>>>>>> REPLACE")[0].strip()
            content = content.replace(search, replace, 1)

        with open(path, "w", encoding="utf-8") as f:
            f.write(content)
        return f"The content of the file {path}  has been replaced successfully."


class SearchFilesTool(Tool):
    """搜索文件工具"""

    def validate(self) -> bool:
        return "path" in self.params and "regex" in self.params

    def execute(self) -> str:
        path = resolve_relative_path(self.cwd, self.params["path"])
        results = []
        for root, _, files in os.walk(path):
            for file in files:
                if "file_pattern" in self.params:
                    if not fnmatch.fnmatch(file, self.params["file_pattern"]):
                        continue

                file_path = os.path.join(root, file)
                with open(file_path, "r", encoding="utf-8") as f:
                    content = f.read()
                    matches = re.finditer(self.params["regex"], content)
                    for match in matches:
                        results.append(f"{file_path}: {match.group()}")

        return "\n".join(results) if results else "No matching content found."


class ListFilesTool(Tool):
    """列出文件工具"""

    def validate(self) -> bool:
        return "path" in self.params

    def execute(self) -> str:
        recursive = str(self.params.get("recursive", "false")).lower() == "true"
        path = resolve_relative_path(self.cwd, self.params["path"])
        return (
            "Listing files completed successfully, result is:\n"
            + get_formatted_filelist_str(path, recursive, 200)
            + "\n\n\n"
        )


class WebAccessTool(Tool):
    """访问网页工具"""

    def validate(self) -> bool:
        return "url" in self.params and "question" in self.params

    def execute(self) -> str:
        url = self.params["url"]
        question = self.params["question"]
        try:
            content = web.get_simplified_webpage(url)[: 32 * 1024]
            result = ai.chat(
                [
                    {
                        "role": "user",
                        "content": f"""Please answer user question based on web page content. The user's question is:
{question}
Web page content is:
{content}""",
                    }
                ],
            )
            return f"Web page access completed, result is: {result}"
        except Exception as e:
            print("WebAccessTool Exception", e)
            return f"Failed to access web page: {str(e)}"


class AskFollowUpQuestionTool(Tool):
    """询问后续问题工具"""

    def validate(self) -> bool:
        return "question" in self.params

    def execute(self) -> str:
        info("assistant", self.params["question"])
        return f"Asked user: {self.params['question']}, Waiting for user replied.\n\n"


class AttemptTaskCompletionTool(Tool):
    """完成所有任务工具"""

    def validate(self) -> bool:
        return "result" in self.params

    def execute(self) -> str:
        result = self.params["result"]
        if (
            "command" in self.params
            and self.params["command"] is not None
            and type(self.params["command"]) == str
            and len(self.params["command"]) > 0
        ):
            result = subprocess.run(
                self.params["command"],
                shell=True,
                capture_output=True,
                text=True,
                cwd=self.cwd,
            )
            return f"Task completed: {result}, executed command: {self.params['command']}, execution result: {result.stdout + result.stderr}"
        else:
            return f"Task completed: {result}!"


class ToolExecutor:
    """工具执行器"""

    TOOL_MAP = {
        "execute_command": ExecuteCommandTool,
        "read_file": ReadFileTool,
        "read_multiple_files": ReadMultipleFilesTool,
        "write_or_overwrite_file": WriteFileTool,
        "replace_part_of_a_file": ReplaceInFileTool,
        "search_files": SearchFilesTool,
        "list_files": ListFilesTool,
        "access_webpage": WebAccessTool,
        "ask_follow_up_question": AskFollowUpQuestionTool,
        "complete_all_tasks": AttemptTaskCompletionTool,
    }

    @classmethod
    def execute_tool(cls, cwd: str, tool_request, abilities) -> Tuple[bool, bool, any]:
        """执行工具"""
        try:
            tool_name = list(tool_request.keys())[0]
            params = tool_request[tool_name]
            if not tool_name:
                return False, False, "Error: Tool type is empty"
            if tool_name in cls.TOOL_MAP:
                tool_class = cls.TOOL_MAP[tool_name]
            elif tool_name in [a.__name__ for a in abilities]:
                func = [a for a in abilities if a.__name__ == tool_name][0]
                try:
                    result = func(**params)
                except Exception as e:
                    print("Execute tool error:", traceback.format_exc())
                    return False, False, traceback.format_exc()
                try:
                    result = json.dumps(result, indent=4, ensure_ascii=False)
                except:
                    return True, False, str(result)
                return True, True, str(result)
            else:
                return (
                    False,
                    False,
                    f"Unknown tool: {tool_name}, the first key of output json ({tool_name}) will be recognized as a tool, so do not output other json except for executing tools.",
                )
            tool = tool_class(params, cwd)
            if not tool.validate():
                return False, False, "Tool parameter validation failed."
            return True, False, tool.execute()
        except Exception as e:
            return False, False, f"Tool execution failed:  {str(e)}"



import json
import os
from .lib import get_formatted_filelist_str, read_local_file, resolve_relative_path
from .message import info
from pycoze.ai import chat_stream_async, extract
from .tools import ToolExecutor
from typing import List


def guess_files_in_message(cwd: str, user_message: str) -> List[str]:

    value = extract(
        {"includedFiles": ["relative path format", "relative path format", "..."]},
        'Please find the files mentioned in the text. If none, return {"includedFiles": []}:\n'
        + user_message,
    )
    return [resolve_relative_path(cwd, p) for p in value["includedFiles"]]


def user_task_prompt(conversation_history, cwd, user_input: str, programmer_mode: bool):
    if programmer_mode:
        potential_paths = guess_files_in_message(cwd, user_input)

        exist_files = get_formatted_filelist_str(cwd, True, 200)
        content = []
        for file_path in potential_paths:
            file_path = resolve_relative_path(cwd, file_path)
            if os.path.isfile(file_path):
                file_marker = f"[[{file_path}]]'"
                file_content = read_local_file(file_path)
                if not any(
                    file_marker in msg["content"] for msg in conversation_history
                ):
                    content.append(f"{file_marker}\n{file_content}")
        content_str = (
            "Partial contents of files are as follows:" + "\n".join(content)
            if content
            else ""
        )
        return f"""
<task>
{user_input}
</task>

<environment_details>
Current working directory: {cwd}

List of files under path:
{exist_files}

{content_str}

</environment_details>
    """
    else:
        return f"""
<task>
{user_input}
</task>
"""


def dumps_markdown_json(data):
    json_str = json.dumps(data, indent=4, ensure_ascii=False)
    return f"\n```json\n{json_str}\n```\n"


def process_code_block(language: str, content: str):
    """
    处理并清理代码块内容，返回格式化后的字符串
    """
    if language and content:
        # 清理代码块内容
        cleaned_content = "\n".join(
            line.strip() for line in content.split("\n") if line.strip()
        )
        return (language, cleaned_content)
    return ""


async def stream_openai_response(conversation_history, start_new_stream):
    """
    异步流式传输 OpenAI 聊天完成响应并处理结构化输出
    """
    stream = None
    buffer = ""
    in_code_block = False
    code_block_language = ""
    code_block_content = ""
    text_content = ""

    while True:
        # 检查是否需要重新创建流
        if stream is None or start_new_stream["value"]:
            if stream is not None:
                await stream.aclose()  # 关闭之前的流
            stream = chat_stream_async(conversation_history)  # 获取新的异步生成器
            start_new_stream["value"] = False  # 重置标志
            buffer = ""
            in_code_block = False
            code_block_language = ""
            code_block_content = ""
            text_content = ""

        # 使用 async for 迭代异步生成器
        try:
            async for chunk in stream:
                info("assistant", chunk)
                buffer += chunk

                # 检查是否需要重新创建流
                if start_new_stream["value"]:
                    break  # 退出当前的 async for 循环，进入下一次 while 循环

                # 处理 buffer 中的每一行
                while "\n" in buffer:
                    line, buffer = buffer.split("\n", 1)
                    if not in_code_block:
                        if line.strip().startswith("```"):
                            if text_content:
                                yield ("text", text_content.strip())
                                text_content = ""
                            in_code_block = True
                            code_block_language = line.strip()[3:].strip()
                        else:
                            text_content += line + "\n"
                    else:
                        if line.strip().startswith("```"):
                            in_code_block = False
                            yield process_code_block(
                                code_block_language, code_block_content
                            )
                            code_block_content = ""
                        else:
                            code_block_content += line + "\n"

            # 如果流正常结束，退出 while 循环
            break

        except Exception as e:
            # 捕获其他异常（如网络错误）
            print(f"Error: {e}", style="bold red")
            break

    # 处理 buffer 中剩余的内容
    if buffer:
        if in_code_block:
            buffer = buffer.split("```")[0]
            code_block_content += buffer + "\n"
            yield process_code_block(code_block_language, code_block_content)
        else:
            text_content += buffer
            if text_content:
                yield ("text", text_content.strip())


async def handle_user_inputs(
    conversation_history, user_input, cwd, abilities, has_any_tool, bot_setting
):
    no_exit_if_incomplete = bot_setting["systemAbility"]["no_exit_if_incomplete"]
    programmer_mode = bot_setting["systemAbility"]["programmer_mode"]

    start_new_stream = {
        "value": False
    }  # 当遇到AI准备执行JSON，即需要新信息的时候，用于强制停止当前stream，减少后续无效的tokens

    print("Processing user command", user_input)
    if user_input.lower() in ["exit", "quit"]:
        exit(0)
    # 将用户消息添加到对话历史
    conversation_history.append(
        {
            "role": "user",
            "content": user_task_prompt(
                conversation_history, cwd, user_input, programmer_mode
            ),
        }
    )
    need_break = False

    if no_exit_if_incomplete:
        okay_str = 'Okay, please continue. If the tasks within <task>...task content...</task> have been completed, execute the tool "complete_all_tasks". If you have a question, use "ask_follow_up_question".'
    else:
        okay_str = "Okay"
    while True:
        async for response in stream_openai_response(
            conversation_history, start_new_stream
        ):
            if len(response) == 2:
                if (
                    response[0] == "text"
                    and response[1].strip() != ""
                    or (response[0] == "json" and not has_any_tool)
                ):
                    conversation_history.append(
                        {"role": "assistant", "content": response[1]}
                    )
                    conversation_history.append(
                        {
                            "role": "user",
                            "content": okay_str,
                        }
                    )
                    continue
                elif response[0] == "json":
                    info("assistant", "\n")
                    cleaned_content = response[1]
                    try:
                        tool_request = json.loads(cleaned_content)
                        tool_name = list(tool_request.keys())[0]
                    except json.JSONDecodeError as e:
                        conversation_history.append(
                            {
                                "role": "assistant",
                                "content": f"\n```json\n{cleaned_content}\n```\n",
                            }
                        )
                        conversation_history.append(
                            {
                                "role": "user",
                                "content": "Invalid JSON content:" + str(e),
                            }
                        )
                        continue

                    ok, is_json_dumps, result = ToolExecutor.execute_tool(
                        cwd, tool_request, abilities
                    )
                    if ok:
                        info("assistant", "✅\n")
                    else:
                        info("assistant", "❌\n")
                    assistant_content = (
                        "Executing tool: "
                        + dumps_markdown_json(tool_request)
                        + "\n\nResult: "
                        + result
                    )
                    if is_json_dumps:
                        info("assistant", "\n```json\n" + result + "\n```\n\n")
                    else:
                        info("assistant", "\n```text\n" + result + "\n```\n\n")
                    conversation_history.append(
                        {"role": "assistant", "content": assistant_content}
                    )
                    if tool_name in ["complete_all_tasks", "ask_follow_up_question"]:
                        need_break = True
                        break
                    else:
                        conversation_history.append(
                            {
                                "role": "user",
                                "content": okay_str,
                            }
                        )
                        start_new_stream["value"] = True

        if need_break:
            break
        if not no_exit_if_incomplete and not start_new_stream["value"]:
            break
    print("task end")


# 示例调用
# user_input_list = [
#     "访问https://api-docs.deepseek.com/zh-cn/guides/chat_prefix_completion，并结合它编写一段代码，并保存"
# ]

# asyncio.run(handle_user_inputs(user_input_list))
